/*
 *      b t r e e c o m _ c
 *
 *  This contains internal support routines that are
 *  common to the implementation of both standard binary 
 *  and balanced binary trees
 *
 *
 * (c) Copyright 1996 David J. Walker
 *     Email:  d.j.walker@slh0101.wins.icl.co.uk
 *
 *     Permission to copy and/or distribute granted under the
 *     following conditions:
 *
 *     1). This notice must remain intact.
 *     2). The author is not responsible for the consequences of use
 *         this software, no matter how awful, even if they
 *         arise from defects in it.
 *     3). Altered version must not be represented as being the
 *         original software.
 *
 *
 *  AMENDMENT HISTORY
 *  ~~~~~~~~~~~~~~~~~
 *  12 May 96   DJW   - First version
 */

#include <liblist.h>
#include <sys/liblist.h>


/*========================================================== BTREE_LEFT */
static
_btree_t    BTREE_Left  (_btree_t node)
/*          ~~~~~~~~~~
 *  This function will loop down left nodes until 
 *  we find a node that is a leaf.
 *----------------------------------------------------------------------*/
{
    while (node->left != NULL)
    {
        node = node->left;
    }
    return (node);
}

/*========================================================= BTREE_RIGHT */
static
_btree_t    BTREE_Right  (_btree_t node)
/*          ~~~~~~~~~~~
 *  This function will loop down right nodes until 
 *  we find a node that is a leaf.
 *----------------------------------------------------------------------*/
{
    while (node->right != NULL)
    {
        node = node->right;
    }
    return (node);
}


/*==================================================== _LIST_BTREEFIRST */
extern
node_t    _LIST_BtreeFirst  (list_t list)
/*        ~~~~~~~~~~~~~~~~
 *  This function will find the first node by
 *  starting at the base and follwing left
 *  branches until a leaf is reached.
 *----------------------------------------------------------------------*/
{
    _btree_t    node = list->nodes;

    if (node != NULL)
    {
        node = BTREE_Left(node);
        assert (node != NULL);
        node = _LIST_USERNODE(list,node);
    }
    return (node);
}


/*==================================================== _LIST_BTREELAST */
extern
node_t    _LIST_BtreeLast  (list_t list)
/*        ~~~~~~~~~~~~~~~
 *  This function will find the last node by
 *  starting at the base and follwing right
 *  branches until a leaf is reached.
 *----------------------------------------------------------------------*/
{
    _btree_t    node = list->nodes;

    if (node != NULL)
    {
        node = BTREE_Right(node);
        assert (node != NULL);
        node = _LIST_USERNODE(list,node);
    }
    return (node);
}


/*================================================== _LIST_BTREEPREVIOUS */
extern
node_t  _LIST_BtreePrevious      (list_t list, node_t node)
/*      ~~~~~~~~~~~~~~~~~~~
 *-----------------------------------------------------------------------*/
#define leaf   ((_btree_t)node)
{
    /*
     *  If there is a left strand to
     *  the current leaf, it will contain
     *  the previous node.
     */
    if (leaf->left)
    {
        leaf = BTREE_Right(leaf->left);
    }
    else
    {
        /*
         *  If there is no left strand then go
         *  upwards until we find one of which
         *  we are not already the left branch.
         */
        for ( ; ; )
        {
            _btree_t   oldleaf;

            oldleaf = leaf;
            leaf = oldleaf->parent;
            /*
             *  If we are already at the top level,
             *  then there cannot be a lesser node.
             */
            if (leaf == NULL)
            {
                return NULL;
            }
            /*
             *  If we came from the right branch,
             *  then this is the node that we want.
             */
            if (leaf->right == oldleaf)
            {
                break;
            }
        }
    }
    leaf = _LIST_USERNODE(list,leaf);
    return (leaf);
}
#undef leaf


/*====================================================== _LIST_BTREENEXT */
extern
node_t  _LIST_BtreeNext      (list_t list, node_t node)
/*      ~~~~~~~~~~~~~~~
 *  To find the next node, we need to either
 *  descend down the right hand path until we get to
 *  a new greatest node, or if there is no right
 *  node we must ascend through the parent links
 *  until we find a node that has a right link
 *  or we run out of nodes.
 *-----------------------------------------------------------------------*/
#define leaf  ((_btree_t)node)
{
    /*
     *  If there is a right strand to
     *  the current leaf, it will contain
     *  the next node.
     */
    if (leaf->right != NULL)
    {
       leaf = BTREE_Left (leaf->right);
    }
    else
    {
        /*
         *  If there is no right strand then
         *  Go upwards until we find one of
         *  which we are not already the 
         *  right branch.
         */
        for ( ; ; )
        {
            _btree_t   oldleaf;

            oldleaf = leaf;
            leaf = oldleaf->parent;
            /*
             *  If we have run out of levels,
             *  then we are at the end of the list.
             */
            if (leaf == NULL)
            {
                return NULL;
            }
            /*
             *  If we came from the left branch,
             *  then this is the node that we want.
             */
            if (leaf->left == oldleaf)
            {
                break;
            }
        }
    }
    leaf = _LIST_USERNODE(list,leaf);
    return (leaf);
}
#undef leaf


/*===================================================== _LIST_BTREEREMOVE */
extern
int     _LIST_BtreeRemove    (list_t list, node_t node)
/*      ~~~~~~~~~~~~~~~~~
 *  Remove a node from the list.
 *
 *  This is easy if it is a leaf node, 
 *  but rather more difficult if it is not.
 *-----------------------------------------------------------------------*/
#define _node ((_btree_t)node)
{   
    register _btree_t  parent;       /* parent node */
    register _btree_t  child;        /* Child node to be promoted a level */

    assert (list != NULL);
    assert (node != NULL);

#ifndef NDEBUG
    if (_node->parent == NULL)
    {
        assert (list->nodes == node);
    }
#endif
    /*
     *  We promote the right leg in
     *  preference to the left leg.
     */
    if (_node->right != NULL)
    {
        child = _node->right;
        /*
         *  Attache nodes old left branch at correct point
         */
        (BTREE_Left(_node->right))->left = _node->left;
    }
    else 
    {
        child = _node->left;
    }
    
    parent = _node->parent;
    if (parent == NULL)
    {
        /*
         *  The old node was the base of the
         *  tree, so we need to set up a new
         *  base to the tree.
         */
        list->nodes = child;
    }
    else
    {
        /*
         *  If the old node was not the base,
         *  then we must adjust the correct
         *  leg of the parent to point to
         *  its new child node.
         */
        if (parent->right == _node)
        {
            parent->right = child;
        }
        else
        {
            assert (parent->left == _node);
            parent->left = child;
        }
    }
    /*
     *  Now connect the child to its new parent
     */
    child->parent = parent;

    /*
     *  Clear any pointers from this node
     *  so we can tell later it is not
     *  attached to any list.
     */
    _node->parent = NULL;
    _node->right = NULL;
    _node->left = NULL;
    /*
     *  Finally correct count
     *  and do a good exit!
     */
    return (0);
}
#undef _node


/*====================================================== _LIST_BTREEFIND */
extern
node_t  _LIST_BtreeFind (list_t list, int (*findfunc)(node_t, va_list), va_list ap)
/*      ~~~~~~~~~~~~~~~
 *  This routine walks down a binary tree to
 *  find the node that matches the desired
 *  value, or until it cannot be found.
 *-----------------------------------------------------------------------*/
{
    register _btree_t listnode;
    register node_t usernode;
    register int flag;
    register int reply;

    /*
     *  If the list is unordered, then
     *  search from the start.  If it is
     *  ordered then we try and start
     *  search (not the best way!).
     */
    for (flag = 0, listnode = list->nodes ; listnode != NULL ; )
    {
        usernode = _LIST_USERNODE(list,listnode);
        reply = (*findfunc)(usernode,ap);

        if (reply == 0) 
        {
            return (usernode);
        }
        switch (flag)
        {
        case 0:     /* Search type unknown */
                if (list->class->compare == NULL || reply == LIST_ORDER_NA)
                {
                    listnode = _LIST_REALNODE(list,LIST_First(list));
                    flag = 2;
                    break;
                }
                flag = 1;
                /* FALLTHRU */
        case 1:     /* Search type ordered */
                listnode = (reply < 0) ? listnode->right : listnode->left;
                break;

        case 2:     /* Search type unordered */
                listnode = _LIST_REALNODE(list, _LIST_BtreeNext(list,listnode));
                break;
        }
    }

    return NULL;
}

