/*
 *      L I S T _ C o m p a r e
 *
 *  This function is called when it is required
 *  to compare two nodes.
 *
 *  If the list is ordered, then we simply use 
 *  the compare routine provided.   If it is not
 *  ordered, then we go through the list to find
 *  which occurs earlier.  We make some special
 *  assumptions if either or both of them are
 *  not on the list:
 *      -ve     node2 not on list
 *      +ve     node1 not on list
 *      0       neither node on list
 *
 *
 * (c) Copyright 1996 David J. Walker
 *     Email:  d.j.walker@slh0101.wins.icl.co.uk
 *
 *     Permission to copy and/or distribute granted under the
 *     following conditions:
 *
 *     1). This notice must remain intact.
 *     2). The author is not responsible for the consequences of use
 *         this software, no matter how awful, even if they
 *         arise from defects in it.
 *     3). Altered version must not be represented as being the
 *         original software.
 *
 *
 *  AMENDMENT HISTORY
 *  ~~~~~~~~~~~~~~~~~
 *  12 May 96   DJW   - First version
 */


#include <liblist.h>
#include <sys/liblist.h>


int     LIST_Compare (list_t list, node_t node1, node_t node2)
{
    node_t  node;

    debug ("LIST_Compare", \
           "'list' parameter is not a valid list instance", \
           _LIST_ListType(list));
    /*
     *  First check for either (or both)
     *  of the node parameters being NULL
     *  as we then take special action
     */
    if (node1 == NULL)
    {
        return (node2 == NULL) ? 0 : 1;
    }
    else if (node2 == NULL)
    {
        return -1;
    }

    debug ("LIST_Compare", \
           "node1 is not a valid node for this list type", \
           _LIST_ListId(list,node1));
    debug ("LIST_Compare", \
           "node2 is not a valid node for this list type", \
           _LIST_ListId(list,node2));
    /*
     *  If a user-defined compare function
     *  is provided, then use that to
     *  do the comparison of the nodes.
     */
    if (list->class->compare != NULL)
    {
        return  (*(list->class->compare))(node1,node2);
    }

    /*
     *  If no-user defined function is provided
     *  then we do a comparison based on whether
     *  the nodes are on the list or not.
     */
    for (node = LIST_First(list) ; node != NULL ;  node = LIST_Next(list,node))
    {
        if (node == node1)
        {
            return -1;
        }
        if (node == node2)
        {
            return 1;
        }
    }
    return 0;
}

