/*
 *      h a s h s h a r _ c
 *
 *  This contains routines that are specific
 *  to the implementation of hash tables, but
 *  that contain no code that would stop their
 *  inclusion into the RLL library variant.
 *
 *
 * (c) Copyright 1996 David J. Walker
 *     Email:  d.j.walker@slh0101.wins.icl.co.uk
 *
 *     Permission to copy and/or distribute granted under the
 *     following conditions:
 *
 *     1). This notice must remain intact.
 *     2). The author is not responsible for the consequences of use
 *         this software, no matter how awful, even if they
 *         arise from defects in it.
 *     3). Altered version must not be represented as being the
 *         original software.
 *
 *
 *  AMENDMENT HISTORY
 *  ~~~~~~~~~~~~~~~~~
 *  12 May 96   DJW   - First version
 */

#include <liblist.h>
#include <sys/liblist.h>
#include <errno.h>


/*========================================================= HASH_FIRST */
extern
node_t    _LIST_HashFirst  (list_t list)
/*        ~~~~~~~~~~~
 *  This function will find the first node by
 *  starting at the base and follwing left
 *  branches until a leaf is reached.
 *----------------------------------------------------------------------*/
{
#if 1
    (void)list;
    return (NULL);
#else
    _btree_t    node = list->nodes;

    if (node != NULL)
    {
        node = HASH_Left(node);
        assert (node != NULL);
        node = _LIST_USERNODE(list,node);
    }
    return (node);
#endif
}


/*========================================================== HASH_LAST */
extern
node_t    _LIST_HashLast  (list_t list)
/*        ~~~~~~~~~~~~~~
 *  This function will find the last node by
 *  working backwards through the buckets.
 *  If there is a linked table (the norm) then
 *  it will call the LIST_Last method on the
 *  linked table.  It will repeat this process
 *  until either it runs out of buckets, or one
 *  of the calls returns a valid node.
 *----------------------------------------------------------------------*/
#define _list ((_hashh_t)list)
{
#if 1
    (void)list;
    return NULL;
#else
    _listclass_t class;
    _listmaster_t master;
    _hash_t    bucket;
    int        x;


    if (list->size == 0) 
    {
        errno = EACCES;
        return (NULL);
    }

    class = list->class;
    master = class->master;

    for (x = list->size, bucket = &list->nodes[list->size - 1] ; 
                x > 0 ; x--)
    {
        /*
         *  If there is nothing in a bucket,
         *  then continue looking.
         */
        if (*(--bucket) == NULL)
        {
            continue;
        }
        /*
         *  If there is no linked table, this
         *  must be the node that we want.
         */
        if (class->linked == NULL)
        {
            break;
        }
    }
    if (x == 0)
    {
        errno = ENOENT;
        return NULL;
    }
        node = HASH_Right(node);
        assert (node != NULL);
        node = _LIST_USERNODE(list,node);
    }
    return (node);
#endif
}
#undef _list


/*======================================================== TREE_PREVIOUS */
extern
node_t  _LIST_HashPrevious      (list_t list, node_t node)
/*      ~~~~~~~~~~~~~~~~~
 *-----------------------------------------------------------------------*/
#define leaf   ((_btree_t)node)
{
#if 1
    (void)list;
    (void)node;
    return NULL;
#else
    /*
     *  If there is a left strand to
     *  the current leaf, it will contain
     *  the previous node.
     */
    if (leaf->left)
    {
        leaf = HASH_Right(leaf->left);
    }
    else
    {
        /*
         *  If there is no left strand then go
         *  upwards until we find one of which
         *  we are not already the left branch.
         */
        for ( ; ; )
        {
            _btree_t   oldleaf;

            oldleaf = leaf;
            leaf = oldleaf->parent;
            /*
             *  If we are already at the top level,
             *  then there cannot be a lesser node.
             */
            if (leaf == NULL)
            {
                return NULL;
            }
            /*
             *  If we came from the right branch,
             *  then this is the node that we want.
             */
            if (leaf->right == oldleaf)
            {
                break;
            }
        }
    }
    leaf = _LIST_USERNODE(list,leaf);
    return (leaf);
#endif
}
#undef leaf


/*============================================================ TREE_NEXT */
extern
node_t  _LIST_HashNext      (list_t list, node_t node)
/*      ~~~~~~~~~~~~~~
 *  To find the next node, we need to either
 *  descend down the right hand path until we get to
 *  a new greatest node, or if there is no right
 *  node we must ascend through the parent links
 *  until we find a node that has a right link
 *  or we run out of nodes.
 *-----------------------------------------------------------------------*/
#define leaf  ((_btree_t)node)
{
#if 1
    (void)list;
    (void)node;
    return NULL;
#else
    /*
     *  If there is a right strand to
     *  the current leaf, it will contain
     *  the next node.
     */
    if (leaf->right != NULL)
    {
       leaf = HASH_Left (leaf->right);
    }
    else
    {
        /*
         *  If there is no right strand then
         *  Go upwards until we find one of
         *  which we are not already the 
         *  right branch.
         */
        for ( ; ; )
        {
            _btree_t   oldleaf;

            oldleaf = leaf;
            leaf = oldleaf->parent;
            /*
             *  If we have run out of levels,
             *  then we are at the end of the list.
             */
            if (leaf == NULL)
            {
                return NULL;
            }
            /*
             *  If we came from the left branch,
             *  then this is the node that we want.
             */
            if (leaf->left == oldleaf)
            {
                break;
            }
        }
    }
    leaf = _LIST_USERNODE(list,leaf);
    return (leaf);
#endif
}
#undef leaf


/*=========================================================== TREE_AFTER */
extern
int     _LIST_HashAfter     (list_t list, node_t newnode, node_t oldnode)
/*      ~~~~~~~~~~~~~~~
 *  Add a node after the given node.
 *-----------------------------------------------------------------------*/
#define _newnode  ((_btree_t)newnode)
#define _oldnode  ((_btree_t)oldnode)
{
#if 1
    (void)list;
    (void)newnode;
    (void)oldnode;
    return (EINVAL);
#else
    if (oldnode == NULL)
    {
        assert (list->nodes == NULL);
        _oldnode = NULL;
    }
    else
    {
        assert (list->nodes != NULL);
        _oldnode = _LIST_REALNODE(list,oldnode);
    }

    HASH_AddNode (list, _newnode, _oldnode);

    return (0);
#endif
}
#undef _newnode
#undef _oldnode


/*=========================================================== HASH_ADD */
extern
int     _LIST_HashAdd       (list_t list, node_t node)
/*      ~~~~~~~~~~~~~
 *  Add a node at the correct point.
 *----------------------------------------------------------------------*/
#define _list ((_btreeh_t)list)
#define _node ((_btree_t)node)
{
#if 1
    (void)list;
    (void)node;
    return EINVAL;
#else
    assert (node != NULL);
    assert (_node->parent == NULL);
    assert (_node->left == NULL);
    assert (_node->right == NULL);

    HASH_AddNode(list, _node, _list->nodes);

    return (0);
#endif
}
#undef _list
#undef _node


/*========================================================== TREE_REMOVE */
extern
int     _LIST_HashRemove    (list_t list, node_t node)
/*      ~~~~~~~~~~~~~~~~
 *  Remove a node from the list.
 *
 *  This is easy if it is a leaf node, 
 *  but rather more difficult if it is not.
 *-----------------------------------------------------------------------*/
#define _node ((_btree_t)node)
{   
#if 1
    (void)list;
    (void)node;
    return EINVAL;
#else
    register _btree_t  parent;       /* parent node */
    register _btree_t  child;        /* Child node to be promoted a level */

    assert (list != NULL);
    assert (node != NULL);
    assert (_LIST_ListId(list,node));

    _node = _LIST_REALNODE(list,node);

#ifndef NDEBUG
    if (_node->parent == NULL)
    {
        assert (list->nodes == node);
    }
#endif /* ! NDEBUG */
    /*
     *  We promote the right leg in
     *  preference to the left leg.
     */
    if (_node->right != NULL)
    {
        child = _node->right;
        /*
         *  Attache nodes old left branch at correct point
         */
        (HASH_Left(_node->right))->left = _node->left;
    }
    else 
    {
        child = _node->left;
    }
    
    parent = _node->parent;
    if (parent == NULL)
    {
        /*
         *  The old node was the base of the
         *  tree, so we need to set up a new
         *  base to the tree.
         */
        list->nodes = child;
    }
    else
    {
        /*
         *  If the old node was not the base,
         *  then we must adjust the correct
         *  leg of the parent to point to
         *  its new child node.
         */
        if (parent->right == _node)
        {
            parent->right = child;
        }
        else
        {
            assert (parent->left == _node);
            parent->left = child;
        }
    }
    /*
     *  Now connect the child to its new parent
     */
    child->parent = parent;

    /*
     *  Clear any pointers from this node
     *  so we can tell later it is not
     *  attached to any list.
     */
    _node->parent = NULL;
    _node->right = NULL;
    _node->left = NULL;
    /*
     *  Finally correct count
     *  and do a good exit!
     */
    return (0);
#endif
}
#undef _node


/*=========================================================== HASH_FIND */
extern
node_t  _LIST_HashFind (list_t list, int (*findfunc)(node_t, va_list), va_list ap)
/*      ~~~~~~~~~~~~~~
 *  This routine walks down a binary tree to
 *  find the node that matches the desired
 *  value, or until it cannot be found.
 *-----------------------------------------------------------------------*/
{
#if 1
    (void) list;
    (void) findfunc;
    (void) ap;
    return (NULL);
#else
    int reply;
    va_list ap;

    /*
     *  Check to see if leaf reached
     */
    if (node == NULL)
    {
        return NULL;
    }

    va_start (ap,list);
    reply = (*findfunc) (lptr, node, ap);
    va_end(ap);
    if (reply != 0)
        _btree_t _node;

        _node = (_btree_t)_LIST_REALNODE(list, node);
        if (reply < 0)
        {
            _node = node->left;
        }
        else
        {
            _node = _node->right;
        }
        node = _LIST_USERNODE(list,_node);

        return HASH_Find (list, (node, ap);

    }
    return (node);
#endif
}

