/*
 *      l i s t 2 _ c
 *
 *  This contains routines that are specific
 *  to the implementation of doubly linked lists
 *
 *
 * (c) Copyright 1996 David J. Walker
 *     Email:  d.j.walker@slh0101.wins.icl.co.uk
 *
 *     Permission to copy and/or distribute granted under the
 *     following conditions:
 *
 *     1). This notice must remain intact.
 *     2). The author is not responsible for the consequences of use
 *         this software, no matter how awful, even if they
 *         arise from defects in it.
 *     3). Altered version must not be represented as being the
 *         original software.
 *
 *
 *  AMENDMENT HISTORY
 *  ~~~~~~~~~~~~~~~~~
 *  12 May 96   DJW   - First version
 */

#include <liblist.h>
#include <sys/liblist.h>

#include <errno.h>
#include <stdio.h>


/*======================================================== LIST2_PREVIOUS */
static 
node_t  LIST2_Previous  (list_t list, node_t node)
/*      ~~~~~~~~~~~~~~
 *------------------------------------------------------------------------*/
#define _node ((_list2_t)node)
{

    _node = _node->prev;

    if (_node != NULL)
    {
        _node = _LIST_USERNODE(list, node);
    }
    return (node);

}
#undef _node


/*========================================================== LIST2_AFTER */
static 
int     LIST2_After     (list_t list, node_t newnode, node_t oldnode)
/*      ~~~~~~~~~~~
 *  Add an entry to a doubly linked list
 *  after the given node.  If the given
 *  node is NULL, then isert at the
 *  start of the list.
 *--------------------------------------------------------------------*/
#define _list    ((_list2h_t)list)
#define _newnode ((_list2_t)newnode)
#define _oldnode ((_list2_t)oldnode)
{
    debug ("LIST2_Remove", \
           "'next' pointer already set in node", \
            _newnode->next == NULL);
    debug ("LIST2_Remove", \
            "'prev' pointer already set in node", \
            _newnode->prev == NULL);

    /*
     * Empty list ?
     */
    if (_list->nodes == NULL)
    {
        assert (_oldnode == NULL);
        assert (_list->last == NULL);

        _list->nodes = _newnode;
        _list->last = _newnode;
    }
    else
    {
        assert (_list->last != NULL);

        if (_oldnode == NULL)
        {
            /*
             *  At start of existing list
             */
            _list->nodes->prev = _newnode;
            _newnode->next = _list->nodes;
            _list->nodes = _newnode;
        }
        else
        {
            /*
             *  Into middle of existing list
             */

            if (_oldnode->next == NULL)
            {
                /*
                 *  If the old node was the end of
                 *  the list, then the new one is now.
                 */
                assert (_list->last == _oldnode);
                _list->last = _newnode;
            }
            else
            {
                /*
                 *  If the old node was in the middle of
                 *  the list, then update next node
                 *  previous pointer.
                 */
                _oldnode->next->prev = _newnode;
            }
            _newnode->next = _oldnode->next;

            _oldnode->next = _newnode;
            _newnode->prev = _oldnode;
        }
    }

    return (0);
}
#undef _list
#undef _newnode
#undef _oldnode


/*========================================================= LIST2_REMOVE */
static 
int     LIST2_Remove    (list_t list, node_t node)
/*      ~~~~~~~~~~~~
 *  Remove the specified entry from a doubly linked list.
 *-----------------------------------------------------------------------*/
#define _list    ((_list2h_t)list)
#define _node ((_list2_t)node)
{
    /*
     *  We had better check for a NULL
     *  parameter as this could cause havoc.
     */
    if (node == NULL)
    {
        return EINVAL;
    }
    debug ("LIST2_Remove", \
           "node is not linked in to a list", \
          (_node->next != NULL || _node->prev != NULL)
        ||(_node == _list->nodes && _node == _list->last) );

    /*
     *  If there is a forward pointer, then
     *  change backward link in next node.
     */
    if (_node->next != NULL)
    {
        _node->next->prev = _node->prev;
    }
    else
    {
        /*
         *  No forward link means this should
         *  be the last node in the list.
         */
        assert (_list->last == _node);
        _list->last = _node->prev;
    }
    /*
     *  If there is a backward pointer, then
     *  change forward link in previous node.
     */
    if (_node->prev != NULL)
    {
        _node->prev->next = _node->next;
    }
    else
    {
        /*
         *  No backward link means this should
         *  be the first node in the list.
         */
        assert (_list->nodes == _node);
        _list->nodes = _node->next;
    }

    /*
     *  Tidy up and exit.
     */
    _node->next = NULL;
    _node->prev = NULL;
    return (0);
}
#undef _list
#undef _node


/*========================================================================
/*
 *  This is the list of vectored functions
 *  that are used to support this list type.
 *
 *------------------------------------------------------------------------*/

static struct _listmaster master =
        {
        sizeof(struct _list2h),     /* Size of the list header */
        sizeof(struct _list2),      /* Size of the node header */
        &_LIST_ListCount,
        &_LIST_ListFirst,
        &_LIST_FifoLast,
        &_LIST_ListNext,
        &LIST2_Previous,
        NULL,
        NULL,
        &_LIST_ListAdd,
        &LIST2_After,
        &LIST2_Remove,
        &_LIST_ListFind,
        _LIST_DOUBLE,
        "DOUBLE",
        "Double Linked List"
        };

/*
 *  This is the default settings for a new
 *  object of this type.
 */
struct _listclass _LISTlist2 =
        {
        &master,                    /* Pointer to list master class */
        0,                          /* Clone count */
        0,                          /* Size of node data */
        NULL,                       /* User node Initialise */
        NULL,                       /* User node Destroy */
        NULL,                       /* user node Compare */
        NULL                        /* Linked list (none) */
        };



