/*
 *      l i s t t e s t
 *      ~~~~~~~~~~~~~~~
 *  This is a test program for the LIBLIST library.
 *  It can also be considered an example of how to
 *  use the various methods and properties.
 *
 *
 * (c) Copyright 1996 David J. Walker
 *     Email:  d.j.walker@slh0101.wins.icl.co.uk
 *
 *     Permission to copy and/or distribute granted under the
 *     following conditions:
 *
 *     1). This notice must remain intact.
 *     2). The author is not responsible for the consequences of use
 *         this software, no matter how awful, even if they
 *         arise from defects in it.
 *     3). Altered version must not be represented as being the
 *         original software.
 *
 *
 *  AMENDMENT HISTORY
 *  ~~~~~~~~~~~~~~~~~
 *  25 May 96   DJW   - First version
 */


#include <liblist.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <qdos.h>

#ifdef DEBUG
char    _prog_name[] = "LibTest_DBG";
#else
char    _prog_name[] = "LibTest";
#endif /* DEBUG */

/*
 *  A definition for the user part of the list nodes
 *  
 *  Note that using a 'typedef' to create a pointer
 *  to this structure type will make later code
 *  with casts much cleaner and easier to read.
 */
typedef struct MYNODE {
        char  * name;
        time_t  when;
    } * MYNODE_t ;

/*=========================================================== INITNODE */

int     initnode (node_t node, va_list ap)
/*
 *  A 'Node initialisation' function.
 *
 *  It is passed the following parameters via 
 *  the 'ap' variable parameter pointer.
 *      char * name;
 *
 *  The function simply uses the strdup() function to 
 *  allocate the memory to hold a copy of the name and
 *  copy it across.  A pointer to the resulting string
 *  is stored in the node.  This can later be used for
 *  comparison and searching purposese.
 *---------------------------------------------------------------------*/
{
    struct MYNODE *mynode;
    char * name;

   /*
    *   To avoid us having to use excessive numbers of casts,
    *   set up a pointer to the user part of the data that
    *   has the correct data type to enable us to access the
    *   structure members using it.
    *
    *   Also get the 'name' paramater from the variable part
    *   of the parameter list.
    */
    mynode = node; 
    name = va_arg(ap,char *); 

    /*
     *  This next part would probably not
     *  normally be present in most programs.
     *  However as this a test program, we
     *  have far more consistency checks
     *  and progress reports than would
     *  normally be the case.
     */

    (void)printf ("  initialising node at $%p (%s)", node, name);

    if (strlen(name) < 10)
    {
        (void)printf (" *** name parameter invalid ($%p) ***\n",name);
        exit(-1);
    }

    /* 
     *  Do the actual work of making a copy
     *  of the name and link the memory
     *  allocated to the node.
     */
    mynode->name = strdup (name);

    /*
     *  Now some additional checks and reports.
     *  Most user programs assume that strdup()
     *  always works although obviously they
     *  probably should not!
     */
    if (mynode->name == NULL)
    {
        (void)printf (" *** strdup() failed (errno=%d) ***\n",errno);
        return (ENOMEM);
    }
    (void)printf ("  (name at $%p)  DONE\n", mynode->name);

    /*
     *  Finally make a good return
     */
    return 0;
}

/*============================================================== KILLNODE */

int     killnode (node_t node)
/*
 *  A 'Node destruction' function.
 *
 *  It must free the memory that is used
 *  to hold the name information.
 *------------------------------------------------------------------------*/
{
    (void) printf ("  killing node at $%p (name=$%p,'%s')", 
                        node, ((MYNODE_t)node)->name, ((MYNODE_t)node)->name);
    free (((MYNODE_t)node)->name);

    (void) printf ("   DONE\n");
    return (0);
}

/*=============================================================== COMPARE */

int     compare (node_t node1, node_t node2)
/*  
 *  A 'Node compare' function.
 *
 *  It compares on the name fields
 *  so that we can order a list 
 *  alphabetically be node name.
 *------------------------------------------------------------------------*/
{
    return strcmp (((MYNODE_t)node1)->name, 
                   ((MYNODE_t)node2)->name);
}

/*============================================================ FINDNODE */

int   findnode (node_t node, va_list ap)
/*
 *  A 'Node find' function.
 *
 *  It is passed the following parameters via 
 *  the 'ap' variable parameter pointer.
 *      char * name;
 *----------------------------------------------------------------------*/
{
    char * name;

    name = va_arg(ap, char *);

    return strcmp(((struct MYNODE *)node)->name, name);
}

/*============================================================ ENUMNODE */

int   enumnode (node_t node, va_list ap)
/*
 *  A 'Node enumeration' function.
 */
{
    struct MYNODE * mynode;

    (void)ap;
    mynode = node;
    (void)printf ("  Node at $%p, name='%s'\n", node, mynode->name);
    return (0);
}

/*======================================================================*/

static  char * nodenames[] = {
            "always first",
            "z 2nd unordered",
            "third unordered",
            "fourth unordered",
            "fifth unordered",
            "sixth unordered",
            "seventh unordered",
            "eight unordered",
            "ninth unordered",
            "tenth (last) unordered",
            NULL        /* End of list */
            };


/*========================================================= TEST_LISTTYPE */

void    test_listtype (int  online, 
                       struct _listclass * listtype,
                       int  (*initfunc)(node_t, va_list),
                       int  (*killfunc)(node_t),
                       int  (*compfunc)(node_t, node_t),
                       char * description)
/*
 *  This function runs a test of one particular list type.
 *  It is called repeatedly to test each list type in turn.
 *  It is designed to test all methods and properties for
 *  a particular list type.
 *------------------------------------------------------------------------*/
{
    node_t  node;
    list_t  master;
    list_t  list;
    char ** pptr;
    int     ireply;
    long    lreply;

    (void)printf ("\nSTARTING TEST (%s)\n\n",description);

    /*
     *  Lets do list creation.  We create a
     *  master copy, and then clone it.
     */
    (void)printf ("Creating new list (%s)\n", description);
    master = LIST_Create(listtype, sizeof (struct MYNODE),
                        initfunc, killfunc, compfunc);
    if (master == NULL)
    {
        (void)printf ("Failed to create list (%s).  errno=%d\n",
                                        description, errno);
        exit (-1);
    }
    (void)printf ("master list created (%s).  Pointer = $%p\n\n", 
                                        description, master);
    list = LIST_Clone (master);
    if (list == NULL)
    {
        (void)printf ("Failed to clone list (%s).  errno=%d\n",
                                        description, errno);
        exit (-1);
    }
    (void)printf ("list cloned (%s).  Pointer = $%p\n\n", description, list);

    /*
     *  Now we add our list of nodes
     *  to this list.
     */
    (void)printf ("adding nodes(%s)\n",description);
    for (pptr = nodenames; *pptr != NULL ; pptr++)
    {
        if (LIST_NewAdd(list, *pptr) == NULL)
        {
            (void)printf ("Failed to add a new node.  errno=%d\n",errno);
            exit (-1);
        }
    }
    (void)printf ("list now populated (%s).\n  Count = %ld, Ordered=0x%p\n", 
                            description, LIST_Count(list), LIST_Ordered(list));
    if (online) {(void)printf ("Press a key to continue\n"); (void)getchar();}

    /*
     *  Having added the lists, lets see what
     *  happens when we enumerate them.  The
     *  list should be enumerated in logical
     *  (alphabetic in this case) order if the
     *  list is ordered, and the physical order
     *  in which we inserted them if it is not.
     */
    (void)printf ("Enumerating list (%s)\n", description);
    lreply = LIST_Enumerate(list,&enumnode);
    if (lreply != 0)
    {
        (void)printf ("Failed while enumerating list.  reply=%ld\n",lreply);
        exit (-1);
    }
    (void)printf ("list enumeration completed\n");
    if (online) {(void)printf ("Press a key to continue\n"); (void)getchar();}

    /*
     *  We now do various calls on our new
     *  list to check they all give the
     *  expected result.  The expected result
     *  needs to be checked against the
     *  enumeration list.
     */
    (void)printf ("testing miscellaneous calls (%s)\n", description);
    node = LIST_First(list);
    (void)printf ("LIST_First(): First node at $%p (%s)\n",
                                node, ((MYNODE_t)node)->name);
    node = LIST_Last(list);
    (void)printf ("LIST_Last(): Last node at $%p (%s)\n",
                                node, ((MYNODE_t)node)->name);
    node = LIST_Find(list, findnode, nodenames[3]);
    if (node == NULL)
    {
        (void)printf ("Find failed.  errno=%d\n", errno);
        exit (-1);
    }
    (void)printf ("LIST_Find(): Node found at $%p (%s)\n",
                                            node, ((MYNODE_t)node)->name);
    ireply = LIST_Index(list,node);
    (void)printf ("LIST_Index(): Node list position is %d\n", ireply);
    node = LIST_Previous(list,node);
    (void)printf ("LIST_Previous(): Previous node at $%p (%s)\n",
                                node, ((MYNODE_t)node)->name);
    node = LIST_Position(list,ireply);
    (void)printf ("LIST_Position(%d):  node at $%p (%s)\n",
                                ireply, node, ((MYNODE_t)node)->name);
    node = LIST_Next(list,node);
    (void)printf ("LIST_Next(): Next node at $%p (%s)\n",
                                node, ((MYNODE_t)node)->name);
    if (online) {(void)printf ("Press a key to continue\n");  (void)getchar();}

    /*
     *  We have finished with this list,
     *  so lets check we destroy it tidily
     */
    (void)printf("Destroying lists(%s)\n", description);
    ireply = LIST_Destroy(list);
    if (ireply != 0)
    {
        (void)printf ("Failure while destroying list.  reply=%d\n", ireply);
        exit (-1);
    }
    (void)printf ("Clone of list destroyed (%s)\n", description);
    ireply = LIST_Destroy(master);
    if (ireply != 0)
    {
        (void)printf ("Failure while destroying master list.  reply=%d\n", ireply);
        exit (-1);
    }
    (void)printf ("Master list destroyed (%s)\n", description);

    (void)printf ("\nCOMPLETED tests (%s)\n",description);
    if (online) {(void)printf ("Press a key to continue\n"); (void)getchar();}
}

/*================================================================ MAIN */

int     main (int argc, char * argv[])
{
    struct {
        struct _listclass *  type;
        int (*initnode)(node_t,va_list);
        int (*killnode)(node_t);
        int (*compnode)(node_t, node_t);
        char *  desc;
    } *ptr, listtypes[] = {
        {LIST_SINGLE, &initnode, &killnode, &compare, "SINGLE linked - ORDERED"},
        {LIST_SINGLE, &initnode, &killnode, NULL, "SINGLE linked - unordered"},
        {LIST_DOUBLE, &initnode, &killnode, &compare, "DOUBLE linked - ORDERED"},
        {LIST_DOUBLE, &initnode, &killnode, NULL, "DOUBLE linked - unordered"},
        {LIST_FIFO,  &initnode, &killnode, NULL, "FIFO (Queue) - unordered"},
        {LIST_LIFO,  &initnode, &killnode, NULL, "LIFO (Stack) - unordered"},
        {LIST_BTREE, &initnode, &killnode, &compare, "BINARY TREE - ORDERED"},
        {LIST_BTREE, &initnode, &killnode, NULL, "BINARY TREE - unordered"},
        {NULL,NULL,NULL,NULL,NULL}
        };

    (void) argc;    /* lets stop compiler warnings */
    (void) argv;    /* about variables not used    */

    (void)printf ("TEST OF LIBLIST LIBRARY\n");
    (void)printf ("~~~~~~~~~~~~~~~~~~~~~~~\n\n");

    for (ptr = listtypes ; ptr->type != NULL ; ptr++)
    {
        test_listtype (argc == 1,
                       ptr->type, 
                       ptr->initnode, 
                       ptr->killnode, 
                       ptr->compnode, 
                       ptr->desc);
    }

    (void)printf ("\nALL TESTS COMPLETED OK");
    (void)printf ("\n~~~~~~~~~~~~~~~~~~~~~~\n\n");
    return (0);
}

